from .viewmayavi import (
    MayaviPointCloudViewer3d, MayaviTriMeshViewer3d, MayaviPointGraphViewer3d,
    MayaviTexturedTriMeshViewer3d, MayaviLandmarkViewer3d,
    MayaviVectorViewer3d, MayaviColouredTriMeshViewer3d)

PointCloudViewer3d = MayaviPointCloudViewer3d
PointGraphViewer3d = MayaviPointGraphViewer3d
TriMeshViewer3d = MayaviTriMeshViewer3d
TexturedTriMeshViewer3d = MayaviTexturedTriMeshViewer3d
ColouredTriMeshViewer3d = MayaviColouredTriMeshViewer3d
LandmarkViewer3d = MayaviLandmarkViewer3d
VectorViewer3d = MayaviVectorViewer3d
